/**
  *  \file lregexp1.cc
  */

#include "lregexp1.h"

static char regerrorstr[100];

void regerror(char* p)
{
    strcpy(regerrorstr, p);
}

/** \class LRegExp
    \brief Lisp Regexp

    This is a compiled regexp as accessible to the Lisp program. */
LRegExp::LRegExp(std::string re)
{
    pregexp = regcomp((char*) re.c_str());
    if (pregexp == 0)
        throw LLispException("regexp-error", std::string(regerrorstr));
}

LRegExp::~LRegExp()
{
    if (pregexp)
        free(pregexp);
}

std::string
LRegExp::toString()
{
    return "#<regexp>";
}

/** \class LRegResult
    \brief Lisp Regexp Match Result

    This is a regexp result as accessible to the Lisp program. */
LRegResult::LRegResult(std::string value, const LRegExp& expr)
    : value(value),
      nmatch(0)
{
    const char* p = value.c_str();
    ok = (regexec(expr.pregexp, (char*) p) != 0);
    if (ok) {
        for (int i = 0; i < MAX_MATCH; ++i) {
            if (expr.pregexp->startp[i] && expr.pregexp->endp[i]) {
                startp[nmatch] = expr.pregexp->startp[i] - p;
                endp[nmatch] = expr.pregexp->endp[i] - p;
                ++nmatch;
            }
        }
    }
}

LRegResult::~LRegResult()
{ }

LPtr
LRegResult::getMatch(int n) const
{
    if (n >= 0 && n < nmatch) {
        return new LString(value.substr(startp[n], endp[n] - startp[n]));
    } else {
        return LPtr();
    }
}

std::string
LRegResult::toString()
{
    return "#<regexp-result>";
}

//- (make-regexp RE)
//- => Compile the extended regexp RE (a string) into a regexp object
//-    and return that. Throws regexp-error on error.
static LPtr
FMakeRegexp(LPtr args)
{
    return new LRegExp(argString(args));
}

//- (match RE STRING)
//- => Match regexp (see (make-regexp)) against STRING. Returns
//-    a regexp-match object or nil.
static LPtr
FMatch(LPtr args)
{
    LPtr re = getArg(args);
    std::string s = argString(args);
    if (LRegExp* e = re.as<LRegExp>()) {
        LRegResult* r = new LRegResult(s, *e);
        if (r->isOK()) {
            return LPtr(r);
        } else {
            delete r;
            return LPtr();
        }
    } else {
        throw LLispException("type-error", "Expected regexp");
    }
}

//- (match-string INDEX REGEXP-MATCH)
//- => Extract matching string from REGEXP-MATCH object obtained
//-    with (match). Returns string, or nil if INDEX is out of
//-    range. INDEX can be 0 to return the whole matching text,
//-    or 1..9 for parenthesized groups.
static LPtr
FMatchString(LPtr args)
{
    int n = argInt(args);
    LPtr match = getArg(args);
    if (!match) {
        return LPtr();
    } else if (LRegResult* r = match.as<LRegResult>()) {
        return r->getMatch(n);
    } else {
        throw LLispException("type-error", "Expected regexp result");
    }
}

void
initRegexp()
{
    LSymbol::define("make-regexp",  new LBuiltin(FMakeRegexp, 1, 1));
    LSymbol::define("match",        new LBuiltin(FMatch, 2, 2));
    LSymbol::define("match-string", new LBuiltin(FMatchString, 2, 2));
}
